# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2012, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains helpers to read barcodes from cairo A1 surfaces.
"""

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

import os
import tempfile
import cairo
import subprocess
from sdaps import image
from sdaps import defs


def read_barcode(surface, matrix, x, y, width, height, btype="CODE128"):
    """Tries to read the barcode at the given position"""
    result = scan(surface, matrix, x, y, width, height, btype)

    if result == None:
      # Try kfill approach
      result = scan(surface, matrix, x, y, width, height, btype, True)

    return result

def scan(surface, matrix, x, y, width, height, btype="CODE128", kfill=False):
    x0, y0 = matrix.transform_point(x, y)
    x1, y1 = matrix.transform_point(x + width, y + height)

    # Bounding box ...
    x = min(x0, x1)
    y = min(y0, y1)
    width = max(x0, x1) - x
    height = max(y0, y1) - y

    x, y, width, height = int(x), int(y), int(width), int(height)
    # Round the width to multiple of 4 pixel, so that the stride will
    # be good ... hopefully
    width = width - width % 4 + 4

    # a1 surface for kfill algorithm
    a1_surface = cairo.ImageSurface(cairo.FORMAT_A1, width, height)

    cr = cairo.Context(a1_surface)
    cr.set_operator(cairo.OPERATOR_SOURCE)
    cr.set_source_surface(surface, -x, -y)
    cr.paint()

    if kfill:
        pxpermm = (matrix[0] + matrix[3]) / 2
        barwidth = pxpermm * defs.code128_barwidth
        barwidth = int(round(barwidth))

        if barwidth <= 3:
            return

        if barwidth > 6:
            barwidth = 6

        image.kfill_modified(a1_surface, barwidth)

    pbm = image.get_pbm(a1_surface)
    tmp = tempfile.mktemp(suffix='.png', prefix='sdaps-zbar-')
    f = open(tmp, 'wb')
    f.write(pbm)
    f.close()

    # Is the /dev/stdin sufficiently portable?
    proc = subprocess.Popen(['zbarimg', '-q', '-Sdisable', '-S%s.enable' % btype.lower(), tmp], stdout=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    os.unlink(tmp)

    # The following can be used to look at the images
    #rgb_surface = cairo.ImageSurface(cairo.FORMAT_RGB24, width, height)
    #cr = cairo.Context(rgb_surface)
    #cr.set_operator(cairo.OPERATOR_SOURCE)
    #cr.set_source_rgba(1, 1, 1, 1)
    #cr.paint()
    #cr.set_operator(cairo.OPERATOR_OVER)
    #cr.set_source_rgba(0, 0, 0, 1)
    #cr.mask_surface(a1_surface)
    #global b_count
    #rgb_surface.write_to_png("/tmp/barcode-%03i.png" % b_count)
    #b_count += 1

    if proc.returncode == 4:
        return None

    assert(proc.returncode == 0)
    barcode = stdout.split(b'\n')[0]
    assert barcode.split(b':', 1)[0].replace(b'-', b'').lower() == btype.lower().encode('ascii')

    return barcode.split(b':', 1)[1].decode('utf-8')

#b_count = 0



# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
The matrix module adds support to find out rotation matrix for a scanned
page. After loading this module you can access the transformation matrices
(cairo.Martrix instances) via two functions:

  :py:meth:`model.sheet.Image.matrix.mm_to_px` and
  :py:meth:`model.sheet.Image.matrix.px_to_mm`

The return value of :py:meth:`mm_to_px` can be used to convert millimeter
values to pixel on the scanned page. :py:meth:`px_to_mm` returns the inverse
matrix to convert on page pixels to millimeter values of the original document.

Both the millimeter and pixel spaces start at the top left corner of the page.

Warning: The correct values are only available after "recognize" has been run!
"""

import cairo

from sdaps import model
from sdaps import surface
from sdaps import image


class Image(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'matrix'
    obj_class = model.sheet.Image

    def mm_to_px(self, fallback=True):
        """Return the matrix to convert mm to pixel space. If no matrix
        information is available and `fallback` is set to `True` then a matrix
        will be calculated from the size of the image."""
        matrix = self.px_to_mm(fallback)
        if matrix is not None:
            matrix.invert()
        return matrix

    def px_to_mm(self, fallback=True):
        """Return the matrix to convert pixel to mm space. If no matrix
        information is available and `fallback` is set to `True` then a matrix
        will be calculated from the size of the image."""
        if self.obj.raw_matrix is not None:
            return cairo.Matrix(*self.obj.raw_matrix)
        elif fallback:
            # Return a dummy matrix ... that maps the image to the page size
            width, height = self.obj.surface.get_size()
            xres, yres = image.get_tiff_resolution(
                self.obj.sheet.survey.path(self.obj.filename),
                self.obj.tiff_page)

            if xres == 0 or yres == 0:
                xres = width / self.obj.sheet.survey.defs.paper_width
                yres = height / self.obj.sheet.survey.defs.paper_height

            scan_width = width / xres
            scan_height = height / yres

            dx = (self.obj.sheet.survey.defs.paper_width - scan_width) / 2.0
            dy = (self.obj.sheet.survey.defs.paper_height - scan_height) / 2.0

            matrix = cairo.Matrix()

            # Center the image
            matrix.translate(dx, dy)
            matrix.scale(1.0 / xres, 1.0 / yres)

            return matrix
        else:
            return None

    def matrix_valid(self):
        """Checks whether the proper transformation matrix is known."""
        return self.obj.raw_matrix != None

    def set_px_to_mm(self, matrix):
        """Set the stored matrix for the image. You need to pass in the pixel
        to mm space conversion matrix. Unsetting can be done by passing `None`.
        """
        if matrix is not None:
            self.obj.raw_matrix = tuple(matrix)
        else:
            self.obj.raw_matrix = None




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2013, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os.path
import errno
import cv2
import numpy as np
from sdaps import image
import cairo
from sdaps import defs
from sdaps import log

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

try:
    import gi
    gi.require_version('Poppler', '0.18')
    from gi.repository import Poppler, Gio
except:
    log.warn(_("Cannot convert PDF files as poppler is not installed or usable!"))

def iter_images_and_pages(images):
    """This function iterates over a images and also the contained pages. As
    OpenCV is not able to handle multipage TIFF files, we use the SDAPS internal
    loading method for those."""

    for filename in images:
        if not os.path.exists(filename):
            raise IOError(errno.ENOENT, _("File does not exist"), filename)

        pages = 1
        is_tiff = False
        is_pdf = False

        try:
            # Check whether this is a TIFF file (ie. try to retrieve the page count)
            pages = image.get_tiff_page_count(filename)
            is_tiff = True
        except AssertionError:
            pass

        if not is_tiff:
            try:
                gfile = Gio.File.new_for_path(filename)
                pdf_doc = Poppler.Document.new_from_gfile(gfile, None, None)
                pages = pdf_doc.get_n_pages()
                is_pdf = True
            except:
                # Either not PDF/damaged or poppler not installed properly
                pass


        for page in range(pages):
            if is_tiff:
                # TIFF pages are zero based
                surf = image.get_rgb24_from_tiff(filename, page, False)

                img = to_opencv(surf)

            elif is_pdf:
                # Try to retrieve a single fullpage image, if that fails, render
                # document at 300dpi.

                THRESH = 10 #pt

                pdfpage = pdf_doc.get_page(page)
                page_width, page_height = pdfpage.get_size()

                images = pdfpage.get_image_mapping()
                if len(images) == 1 and (
                        abs(images[0].area.x1) < THRESH and
                        abs(images[0].area.y1) < THRESH and
                        abs(images[0].area.x2 - page_width) < THRESH and
                        abs(images[0].area.y2 - page_height) < THRESH):
                    # Assume one full page image, and simply use that.
                    surf = pdfpage.get_image(images[0].image_id)

                else:
                    dpi = 0
                    # Try to detect the DPI of the scan
                    for img in images:
                        if img.area.y2 - img.area.y1 < page_height / 2:
                            continue

                        surf = pdfpage.get_image(img.image_id)
                        # Calculate DPI from height
                        dpi_x = round(surf.get_height() / (img.area.y2 - img.area.y1) * 72)
                        dpi_y = round(surf.get_width() / (img.area.x2 - img.area.x1) * 72)
                        if abs(dpi_x - dpi_y) <= 1:
                            dpi = max(dpi, dpi_x, dpi_y)

                    # Fall back to 300dpi for odd values
                    if dpi < 199 or dpi > 601:
                        dpi = 300

                    surf = cairo.ImageSurface(cairo.FORMAT_RGB24, int(dpi / 72 * page_width), int(dpi / 72 * page_height))
                    cr = cairo.Context(surf)
                    cr.scale(dpi / 72, dpi / 72)
                    cr.set_source_rgb(1, 1, 1)
                    cr.paint()

                    pdfpage.render_for_printing(cr)

                    del cr

                img = to_opencv(surf)

            else:
                img = cv2.imread(filename)

            yield img, filename, page

def sharpen(img):
    blured = cv2.GaussianBlur(img, (0,0), 5)
    img = cv2.addWeighted(img, 1.5, blured, -0.5, 0)

    return img

def to_opencv(surf):
    width = surf.get_width()
    height = surf.get_height()
    stride = surf.get_stride()

    # We need to ensure a sane stride!
    np_width = stride // 4

    # This converts by doing a copy; first create target numpy array
    # We need a dummy alpha channel ...
    target = np.empty((height, np_width), dtype=np.uint32)

    tmp_surf = cairo.ImageSurface.create_for_data(target.data, cairo.FORMAT_RGB24, width, height, stride)
    cr = cairo.Context(tmp_surf)
    # Handle A1 surfaces?
    cr.set_source_surface(surf)
    cr.paint()
    del cr
    tmp_surf.flush()
    del tmp_surf

    # Now, we need a bit of reshaping
    img = np.empty((height, width, 3), dtype=np.uint8)

    # order should be BGR
    img[:,:,2] = 0xff & (target[:,:] >> 16)
    img[:,:,1] = 0xff & (target[:,:] >> 8)
    img[:,:,0] = 0xff & target[:,:]

    return img

def to_a1_surf(img):
    # Assume that x is the second position
    assert(len(img.strides) == 2)
    assert(img.strides[1] == 1)

    if img.strides[0] % 4:
        # Need to reshape, lets just cut off up to 3 pixel
        target = np.empty((img.shape[0], img.shape[1] - img.shape[1] % 4), dtype=np.uint8)
        target[:,:] = img[:,:target.shape[1]]

        img = target

    height, width = img.shape
    surf_a8 = cairo.ImageSurface.create_for_data(img.data, cairo.FORMAT_A8, width, height, img.strides[0])

    surf_a1 = cairo.ImageSurface(cairo.FORMAT_A1, width, height)
    cr = cairo.Context(surf_a1)
    cr.set_operator(cairo.OPERATOR_SOURCE)
    cr.set_source_surface(surf_a8)
    cr.paint()

    return surf_a1

def ensure_orientation(img, portrait):
    img_portrait = img.shape[1] <= img.shape[0]
    if img_portrait != portrait:
        # Rotate into new array (CV does not like negative strides and such)
        new = np.empty((img.shape[1],img.shape[0])+img.shape[2:], dtype=img.dtype)
        new[:] = np.rot90(img)
        img = new

    return img

def ensure_greyscale(img):
    if len(img.shape) == 2:
        # Well, seems to be greyscale/monochrome already
        return img

    # Average the color samples, and convert back to uint8
    img = np.average(img, 2)
    img = np.array(img, dtype=np.uint8)

    return img

def convert_to_monochrome(img, size=201, thresh_adjust=5):
    img = ensure_greyscale(img)

    img = cv2.adaptiveThreshold(img, 255.0, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY_INV, size, thresh_adjust)

    return img
   
def save(img, filename):
    cv2.imwrite(img, filename)


def _fallback_matrix(width, height, paper_width, paper_height):
    """Calcualte a fallback matrix. Basically the same as in matrix.py, should
    be merged somehow!"""
    xres = width / float(paper_width)
    yres = height / float(paper_height)

    # Assume the smaller size fits better ...
    res = min(xres, yres)

    scan_width = width / res
    scan_height = height / res

    dx = (paper_width - scan_width) / 2.0
    dy = (paper_height - scan_height) / 2.0

    matrix = cairo.Matrix()

    # Center the image
    matrix.translate(dx, dy)
    matrix.scale(1.0 / res, 1.0 / res)

    matrix.invert()

    return matrix, res


def transform_using_corners(img, paper_width, paper_height):
    surf = to_a1_surf(convert_to_monochrome(img))

    matrix, res = _fallback_matrix(surf.get_width(), surf.get_height(), paper_width, paper_height)

    top_left = image.find_corner_marker(surf, matrix, 1)
    top_right = image.find_corner_marker(surf, matrix, 2)
    bottom_right = image.find_corner_marker(surf, matrix, 3)
    bottom_left = image.find_corner_marker(surf, matrix, 4)

    scale = 1 * res

    x0 = scale * defs.corner_mark_left
    y0 = scale * defs.corner_mark_top
    x1 = scale * (paper_width - defs.corner_mark_right)
    y1 = scale * (paper_height - defs.corner_mark_bottom)

    width, height = int(scale * paper_width), int(scale * paper_height)
    # Increase width to be a multiple of 4
    if width % 4:
        width = width + 4 - width % 4

    transform_matrix = cv2.getPerspectiveTransform(
        np.array((top_left, top_right, bottom_right, bottom_left), dtype=np.float32),
        np.array(((x0, y0), (x1, y0), (x1, y1), (x0, y1)), dtype=np.float32))

    transformed = cv2.warpPerspective(img, transform_matrix, dsize=(width, height))

    return transformed




#!/usr/bin/env python3
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright (C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright (C) 2008-2013, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from distutils.core import setup
from distutils.extension import Extension
import glob
import os
import os.path
import subprocess
import sys
from distutils.command import build
from DistUtilsExtra.command import *
import configparser

# We import sdaps to grab the version number; a bit of a hack but
# it should work just fine as few modules are actually loaded by doing
# this
from sdaps import __version__

def pkgconfig(*packages, **kw):
    flag_map = {'-I': 'include_dirs', '-L': 'library_dirs', '-l': 'libraries', '-D' : 'define_macros'}
    (status, tokens) = subprocess.getstatusoutput("pkg-config --libs --cflags %s" % ' '.join(packages))
    if status != 0:
        print(tokens)
        sys.exit(1)

    for token in tokens.split():
        type = flag_map.get(token[:2])
        value = token[2:]
        if type == 'define_macros':
            value = tuple(value.split('=', 1))
        if type is None:
           value = token
           type = 'extra_compile_args'
        kw.setdefault(type, []).append(value)
    return kw

class sdaps_build_tex(build.build):

    description = "build and install the LaTeX packages and classes"

    # Hardcoded ...
    tex_installdir = 'share/sdaps/tex'
    tex_resultdir = 'tex/class/build/local'

    def run(self):
        # Build the LaTeX packages and classes, note that they cannot build
        # out of tree currently.
        maindir = os.path.abspath(os.curdir)
        if not os.path.exists('tex/class/build.lua'):
            print('error: LaTeX build script is not available')
            print('Did you forget to checkout the git submodule? See README for more information.')
            os._exit(1)
        os.chdir('tex/class')
        self.spawn(['./build.lua', 'unpack'])
        os.chdir(maindir)

        files = [os.path.join(self.tex_resultdir, f) for f in os.listdir(self.tex_resultdir)]
        self.distribution.data_files.append((self.tex_installdir, files))


class sdaps_build_i18n(build_i18n.build_i18n):

    # Hardcoded ...
    dict_sourcefile = 'tex/tex_translations.in'
    dict_dir = 'share/sdaps/tex'
    dict_filename = 'tex_translations'

    def run(self):
        # run the original code
        build_i18n.build_i18n.run(self)

        dest_dir = os.path.join('build', self.dict_dir)
        tex_translations = os.path.join(dest_dir, self.dict_filename)

        # Build the tex_translations file
        if not os.path.isdir(dest_dir):
            os.makedirs(dest_dir)
        cmd = ['intltool-merge', '-d', 'tex/po', self.dict_sourcefile, tex_translations]
        self.spawn(cmd)

        ###################
        # Now build the LaTeX dictionaries
        def extract_key_lang(key):
            if not key.endswith(']'):
                return key, None
            index = key.rfind('[')
            return key[:index], key[index+1:-1]

        parser = configparser.ConfigParser()
        parser.read(tex_translations)

        langs = {}
        keys = set()
        for k, v in parser.items("translations"):
            key, lang = extract_key_lang(k)
            if not key == 'tex-language':
                keys.add(key)
                continue

            assert lang not in langs
            assert v not in list(langs.items())

            langs[lang] = v

        # Load mapping from unicode to LaTeX command name
        from sdaps.utils.latex import unicode_to_latex

        dictfiles = []
        for lang, name in langs.items():
            print('building LaTeX dictionary file for language %s (%s)' % (name, lang if lang else 'C'))
            dictfiles.append(os.path.join(dest_dir, 'translator-sdaps-dictionary-%s.dict' % name))
            f = open(dictfiles[-1], 'w')

            f.write('% This file is auto-generated from gettext translations (.po files).\n')
            f.write('% The header of the original file follows for reference:\n')
            f.write('%\n')
            for line in open(self.dict_sourcefile).readlines():
                if not line.startswith('#'):
                    break
                f.write('%' + line[1:])
            f.write('%\n\n')
            f.write('\\ProvidesDictionary{translator-sdaps-dictionary}{%s}\n\n' % name)

            for key in sorted(keys):
                if lang is not None:
                    k = "%s[%s]" % (key, lang)
                else:
                    k = key

                try:
                    value = parser.get("translations", k)
                except configparser.NoOptionError:
                    value = parser.get("translations", key)

                value = unicode_to_latex(value)

                f.write('\\providetranslation{%s}{%s}\n' % (key, value))

        # And install the dictionary files
        self.distribution.data_files.append((self.dict_dir, dictfiles))

class sdaps_clean_i18n(clean_i18n.clean_i18n):
    dict_dir = 'share/sdaps/tex'
    dict_filename = "tex_translations"

    def run(self):
        # Remove dictionaries

        directory = os.path.join('build', self.dict_dir)
        if os.path.isdir(directory):
            print("removing all LaTeX dictionaries in '%s'" % directory)
            for filename in os.listdir(directory):
                if filename.startswith('translator-sdaps-dictionary-'):
                    os.unlink(os.path.join(directory, filename))

        fn = os.path.join('build', self.dict_dir, self.dict_filename)
        if os.path.exists(fn):
            os.unlink(fn)

        clean_i18n.clean_i18n.run(self)

class sdaps_build(build_extra.build_extra):
    sub_commands = build_extra.build_extra.sub_commands + [('build_tex', lambda x : True)]

setup(name='sdaps',
      version=__version__,
      description='Scripts for data acquisition with paper-based surveys',
      url='http://sdaps.sipsolutions.net',
      author='Benjamin Berg, Christoph Simon',
      author_email='benjamin@sipsolutions.net, post@christoph-simon.eu',
      license='GPL-3',
      long_description="""
SDAPS is a tool to carry out paper based surveys. You can create machine
readable questionnaires using LaTeX. It also provides the tools to later
analyse the scanned data, and create a report.
""",
      packages=['sdaps',
                'sdaps.add',
                'sdaps.annotate',
                'sdaps.boxgallery',
                'sdaps.cmdline',
                'sdaps.cover',
                'sdaps.convert',
                'sdaps.csvdata',
                'sdaps.gui',
                'sdaps.image',
                'sdaps.model',
                'sdaps.recognize',
                'sdaps.reorder',
                'sdaps.stamp',
                'sdaps.report',
                'sdaps.reporttex',
                'sdaps.reset',
                'sdaps.setup',
                'sdaps.setuptex',
                'sdaps.utils'
      ],
      package_dir={'sdaps.gui': 'sdaps/gui'},
      scripts=[
               'bin/sdaps',
               ],
      ext_modules=[Extension('sdaps.image.image',
                   ['sdaps/image/wrap_image.c', 'sdaps/image/image.c', 'sdaps/image/transform.c', 'sdaps/image/surface.c'],
                   **pkgconfig('py3cairo', 'cairo', 'glib-2.0', libraries=['tiff']))],
      data_files=[
                  ('share/sdaps/ui',
                   glob.glob("sdaps/gui/*.ui")
                  ),
                  ('share/sdaps/tex', glob.glob('tex/*.cls')
                  ),
                  ('share/sdaps/tex', glob.glob('tex/*.tex')
                  ),
                  ('share/sdaps/tex', glob.glob('tex/*.sty')
                  ),
                  ],
      cmdclass = { "build" : sdaps_build,
                   "build_tex" : sdaps_build_tex,
                   "build_i18n" :  sdaps_build_i18n,
                   "build_help" :  build_help.build_help,
                   "build_icons" :  build_icons.build_icons,
                   "clean" : sdaps_clean_i18n }
     )



# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This modules contains a script for the user to output all boxes sorted by their
coverage. This can be used to adjust magic values for checkbox recognition.
"""

from sdaps import model
from sdaps import script

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

parser = script.add_project_subparser("boxgallery",
    help=_("Create PDFs with boxes sorted by the detection heuristics."),
    description=_("""SDAPS uses multiple heuristics to detect determine the
    state of checkboxes. There is a list for each heuristic giving the expected
    state and the quality of the value (see defs.py). Using this command a PDF
    will be created for each of the heuristics so that one can adjust the
    values."""))

parser.add_argument('--debugrecognition',
    action="store_true",
    help=_('Reruns part of the recognition process and retrieves debug images from this step.'),
    default=False)

@script.connect(parser)
@script.logfile
def boxgallery(cmdline):

    # We need to load the buddies before the survey is loaded.
    if cmdline['debugrecognition']:
        from sdaps.recognize import buddies

    survey = model.survey.Survey.load(cmdline['project'])
    from sdaps import boxgallery
    return boxgallery.boxgallery(survey, cmdline['debugrecognition'])





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
This module alters sys.stdout and sys.stderr to copy all output into a logfile.

Note that some magic is done when the output stream is not a tty. In that
case the progress bar and messages done using the :py:meth:`interactive`
function will be suppressed. The reason to do this is to allow commands to
output data to `sys.stdout`.
'''

import sys
import time
import io

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext


def warn(msg):
    sys.stderr.write(_('Warning: ') + msg + '\n')


def error(msg):
    sys.stderr.write(_('Error: ') + msg + '\n')

def interactive(msg):
    """Only forward the message to stdout if it is a terminal."""
    if hasattr(sys.stdout, "interactive"):
        sys.stdout.interactive(msg)
    else:
        if sys.stdout.isatty():
            sys.stdout.write(msg)

class Copier(object):
    '''copy all data going through the pipe into a logfile'''

    def __init__(self, pipe, logfile):
        self.pipe = pipe
        self.logfile = logfile

    def write(self, data):
        self.pipe.write(data)
        self.logfile.write(data)

    def interactive(self, data):
        if self.pipe.isatty():
            self.pipe.write(data)
        self.logfile.write(data)

    def isatty(self):
        return self.pipe.isatty()

    def flush(self):
        self.pipe.flush()
        self.logfile.flush()

    def fileno(self):
        return self.pipe.fileno()

class Wiper(object):
    '''wipe out the progressbar before forwarding data through the pipe'''

    def __init__(self, pipe, progressbar):
        self.pipe = pipe
        self.progressbar = progressbar

    def write(self, data):
        if self.progressbar.visible:
            self.pipe.write(' ' * 80)
            self.pipe.write('\r')
            self.progressbar.visible = 0
        self.pipe.write(data)

    def isatty(self):
        return self.pipe.isatty()

    def flush(self):
        self.pipe.flush()

    def fileno(self):
        return self.pipe.fileno()

class Encoder(object):
    '''Encode data going through the pipe to utf-8'''

    def __init__(self, pipe):
        self.pipe = pipe

    def write(self, data):
        if isinstance(data, bytes):
            self.pipe.write(data.encode('utf-8'))
        else:
            self.pipe.write(data)

    def close(self):
        self.pipe.close()

    def isatty(self):
        return self.pipe.isatty()

    def flush(self):
        self.pipe.flush()

    def fileno(self):
        return self.pipe.fileno()

class Logfile(object):

    def __init__(self):
        self.logfile = io.StringIO()

    def write(self, data):
        self.logfile.write(data)
        self.logfile.flush()

    def open(self, filename):
        logfile = Encoder(open(filename, 'a'))
        logfile.write(self.logfile.getvalue())
        self.logfile = logfile

    def close(self):
        self.logfile.close()
        self.logfile = io.StringIO()

    def isatty(self):
        return False

    def flush(self):
        self.logfile.flush()

    def fileno(self):
        return self.logfile.fileno()

class ProgressBar(object):

    def __init__(self, pipe):
        self.pipe = pipe
        self.visible = 0

    def start(self, max_value):
        self.max_value = max_value
        self.start_time = time.time()
        self.update(0)

    def update(self, value):
        self.elapsed_time = time.time() - self.start_time

        # Don't display anything if the output is a pipe
        if not self.pipe.isatty():
            return

        progress = float(value) / float(self.max_value)
        self.pipe.write('|')
        self.pipe.write(('#' * int(round(progress * 64))).ljust(64))
        self.pipe.write('| ')
        self.pipe.write(('%i%% ' % int(round(progress * 100))).rjust(5))
        if progress == 0:
            self.pipe.write('--:--:--')
        elif progress == 1:
            self.pipe.write(time.strftime('%H:%M:%S', time.gmtime(self.elapsed_time)))
            self.pipe.write('\n')
        else:
            remaining_time = self.elapsed_time * (1 / progress - 1)
            self.pipe.write(time.strftime('%H:%M:%S', time.gmtime(remaining_time)))
        self.pipe.write('\r')
        self.pipe.flush()
        self.visible = 1

    def isatty(self):
        return self.pipe.isatty()

    def flush(self):
        self.pipe.flush()

progressbar = ProgressBar(sys.stdout)
logfile = Logfile()

redirects_activated = False
def activate_redirects():
    global redirects_activated

    if redirects_activated is not False:
        return
    redirects_activated = True

    sys.stdout = Encoder(sys.stdout)
    sys.stderr = Encoder(sys.stderr)
    sys.stdout = Wiper(sys.stdout, progressbar)
    sys.stderr = Wiper(sys.stderr, progressbar)
    sys.stdout = Copier(sys.stdout, logfile)
    sys.stderr = Copier(sys.stderr, logfile)




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, 2011, 2014, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Ensure the matrix buddy is loaded
from sdaps import matrix

from sdaps import image

import cairo
import math

import os
import os.path

def get_box_surface(img, filename, x, y, width, height, format=cairo.FORMAT_RGB24):
    mm_to_px = img.matrix.mm_to_px()
    x0, y0 = mm_to_px.transform_point(x, y)
    x1, y1 = mm_to_px.transform_point(x + width, y)
    x2, y2 = mm_to_px.transform_point(x, y + height)
    x3, y3 = mm_to_px.transform_point(x + width, y + height)

    x = int(min(x0, x1, x2, x3))
    y = int(min(y0, y1, y2, y3))
    width = int(math.ceil(max(x0, x1, x2, x3) - x))
    height = int(math.ceil(max(y0, y1, y2, y3) - y))

    img = image.get_a1_from_tiff(filename, img.tiff_page, img.rotated if img.rotated else False)

    # Not sure if this is correct for A1 ... probably not
    assert(format == cairo.FORMAT_RGB24)
    surf = cairo.ImageSurface(format, width, height)
    cr = cairo.Context(surf)
    cr.set_source_rgb(1, 1, 1)
    cr.paint()

    cr.set_source_surface(img, -x, -y)
    cr.paint()

    return surf

class ImageWriter:
    def __init__(self, path, prefix):
        # I am not qutie sure whether I like the params here.
        self.count = 0
        self.path = path
        self.prefix = prefix

        # Create directory if it does not exists. Assumes only one level
        # is missing.
        # In case the prefix contains further path components
        real_path = os.path.dirname(os.path.join(self.path, self.prefix))
        if not os.path.exists(real_path):
            os.mkdir(real_path)

    def output_area(self, img, filename, x, y, width, height):
        surf = get_box_surface(img, filename, x, y, width, height)

        filename = "%s%04d.png" % (self.prefix, self.count)
        full_path = os.path.join(self.path, filename)

        surf.write_to_png(full_path)

        self.count += 1

        return filename

    def output_box(self, box):
        img = box.sheet.get_page_image(box.question.page_number)

        filename = box.question.questionnaire.survey.path(img.filename)

        return self.output_area(img, filename, box.data.x, box.data.y, box.data.width, box.data.height)

    def output_boxes(self, boxes, real=False, padding=5):
        x1 = 99999
        y1 = 99999
        x2 = 0
        y2 = 0
        img = boxes[0].sheet.get_page_image(boxes[0].question.page_number)
        filename = boxes[0].question.questionnaire.survey.path(img.filename)

        for box in boxes:
            if real:
                pos = box.data
            else:
                pos = box
            x1 = min(x1, pos.x)
            y1 = min(y1, pos.y)

            x2 = max(x2, pos.x + pos.width)
            y2 = max(y2, pos.y + pos.height)

        x1 -= padding
        y1 -= padding
        x2 += padding
        y2 += padding

        return self.output_area(img, filename, x1, y1, x2 - x1, y2 - y1)


